local foundDebugModule, mobdebug = pcall( require, 'mobdebug' )
local isDebugging = false
if foundDebugModule and mobdebug.start() then
	mobdebug.coro()
	isDebugging = true
end

-- global threads table
_threads = {}
_workingThreads = {}
_worldTime = 0
_screenTime = 0
_eventListeners = {}
_events = {}
_threadStack = nil
_activeThread = nil

-- syntactic sugar for yield
function wait( duration, tag )
	coroutine.yield({ wait = duration, tag = tag })
end

function waitScreenTime( duration, tag )
	coroutine.yield({ wait = duration, screenTime = true, tag = tag })
end

function waitUntil( event )
	-- no need to wait, event already happened
	if _events[event] ~= nil then
		_events[event] = nil
		return
	end

	coroutine.yield({ wait = -1, event = event })
end

function ToLookup( table )
	local lookup = {}
	for key,value in pairs( table ) do
		lookup[value] = true
	end
	return lookup
end

local function EndsWith( String, End )
	return End == '' or string.sub( String, -string.len(End) ) == End
end

DebugFunctionIgnores = ToLookup({
		"dispatch", "resume", "func", "push", "pop", "peek", "newvar",
		"rawset", "for iterator", "pairs", "yield", "on", "isrunning", "create"
	})

MainFileFunctions = ToLookup({ "__newindex", "wait", "waitScreenTime", "waitUntil", "notify", "notifyExistingWaiters", "thread" })


--debugging
function newFunctionCall( triggerType )

	if triggerType == "return" then
		functionReturn()
		return
	end

	local info = nil
	local level = 3
	if isDebugging then
		info = debug.getinfo( level, "nSl" )
	end

	if info == nil then
		level = 2
		info = debug.getinfo( level, "nSl" )
	end

	local threadInfo = NonMainThreadInfo( _activeThread, MainFileFunctions[info.name] == nil )
	if threadInfo ~= nil then
		lastGoodThreadInfo = threadInfo
	end

	if lastGoodThreadInfo ~= nil then
		info.short_src = lastGoodThreadInfo.short_src
		info.currentline = lastGoodThreadInfo.currentline
	end

	if DebugFunctionIgnores[info.name] ~= nil then
		return
	end

	-- dispatch resume push func pop peek

	if info.name == nil or info.linedefined < 0 then
		return
	end

	local text = ""

	if info.name == "__newindex" then

		local name1, newIndexName = debug.getlocal(level, 2)
		local name2, newIndexValue = debug.getlocal(level, 3)
		local valueString = newIndexValue
		if valueString == nil then
			valueString = "nil"
		elseif type(valueString) == "table" then
			--valueString = pickle(valueString)
			--valueString = string.gsub( valueString, "%\n", " " )
			-- @experiment Do not print complete tables on assignment to save time and log space
			valueString = tostring(valueString)
		else
			valueString = tostring(valueString)
		end

		text = text..tostring(newIndexName).." = "..valueString

	else

		text = text..info.name.."("

		-- get parameters
		local a = 1
		local firstParam = false
		while true do
			local name, value = debug.getlocal(level, a)
			if not name then
				break
			end

			if value ~= nil and name ~= "(*temporary)" then
				if type(value) == "string" or type(value) == "number" or type(value) == "boolean" or type(value) == "function" or type(value) == "table" then

					if firstParam then
						text = text..","
					end

					if type(value) == "function" then
						value = findGlobal(value)
					elseif type(value) == "table" and value.ToString ~= nil then
						value = value.ToString( value )
					end

					text = text.." "..name.." = "..tostring(value)
					firstParam = true
				end
			end
			a = a + 1
		end

		text = text.." )"

	end

	local debugOverheadLevels = level + 2
	local stackLevel = GetStackLevel() - debugOverheadLevels
	for i = 1, stackLevel - 1, 1 do
		text = "    "..text
	end
	--text = stackLevel..text

	DebugMessage({ Text = text, FileName = info.short_src, LineNum = info.currentline })

end

function functionReturn()

	local level = 3
	if isDebugging then
		level = 4
	end

	local info = debug.getinfo( level, "nSl" )
	if info == nil or info.name == "__newindex" or info.short_src == "[C]" then
		return
	end

	local localIndex = 1
	while true do
		local name, value = debug.getlocal( level, localIndex )

		if name == "(*temporary)" then
			local valueString = nil
			if value == nil then
				valueString = "nil"
			elseif type(value) == "thread" or value == "done" then
				-- Ignore
				valueString = nil
			elseif type(value) == "table" then
				-- Ignoring table returns for now
				valueString = nil
				--valueString = "table"
				--valueString = pickle(valueString)
				--valueString = string.gsub( valueString, "%\n", " " )
			elseif MainFileFunctions[value] then
				-- Ignore
				valueString = nil
			else
				valueString = tostring(value)
			end
			if valueString ~= nil then
				local text = "Returned: "..valueString
				local debugOverheadLevels = level + 1
				local stackLevel = GetStackLevel() - debugOverheadLevels
				for i = 1, stackLevel - 1, 1 do
					text = "    "..text
				end
				DebugMessage({ Text = text, FileName = info.short_src, LineNum = info.currentline })
				break
			end
		elseif name == nil then
			break
		end

		localIndex = localIndex + 1
	end

end

function StackBottom()

	local level = 1
	local prevInfo = nil

	while true do

		local info = debug.getinfo(level, "Sl")

		if info == nil then
			return prevInfo
		else
			prevInfo = info
		end

		level = level + 1

	end

end

function GetStackLevel()

	local level = 1
	while true do
		local info = debug.getinfo(level, "Sl")
		if info == nil then
			break
		end
		level = level + 1
	end

	return level

end

function ThreadStackBottom( coroutine )

	if coroutine == nil then
		return
	end

	local level = 1
	local prevInfo = nil

	while true do

		local info = debug.getinfo( coroutine, level, "Sl")

		if info == nil then
			return prevInfo
		else
			prevInfo = info
		end

		level = level + 1

	end

end

-- Return info from the highest stack level not in Main.lua
function NonMainThreadInfo( coroutine, plusOneStackLevel )

	if coroutine == nil then
		return
	end

	local level = 1
	local prevInfo = nil
	local returnNextInfo = false

	while true do

		local info = debug.getinfo( coroutine, level, "nSl" )

		if info == nil then
			return nil
		end

		if returnNextInfo then
			return info
		end

		local src = info.short_src
		if not EndsWith( src, "Main.lua" ) and not EndsWith( src, "mobdebug.lua" ) then

			if plusOneStackLevel then
				returnNextInfo = true
			else
				return info
			end
		end

		level = level + 1

	end

end

function SetCurrentLine()

	local threadInfo = nil
	if _threadStack == nil or #_threadStack == 0 then
		threadInfo = debug.getinfo( 1, "Sl" )
		_currentStackLevel = GetStackLevel()
	else
		threadInfo = debug.getinfo( _activeThread, 1, "Sl" )
		_currentStackLevel = 1
		while true do
			local nextLevelInfo = debug.getinfo( _activeThread, _currentStackLevel + 1, "Sl" )
			if nextLevelInfo == nil then
				break
			end
			_currentStackLevel = _currentStackLevel + 1
		end
	end

	if threadInfo == nil then
		_currentLine = 0
		_currentFileName = ""
		return
	end

	_currentLine = threadInfo.currentline
	_currentFileName = threadInfo.short_src

end

function findGlobal( f )
	for k, v in pairs( _G ) do
		if v == f then
			return k
		end
	end
end

function trackNewVariables()
	local mt = {
		__newindex = function( t,k,v )
			newvar( k, v )
			rawset( t,k,v )
		end
	}
	setmetatable( _G, mt )
end

function resume( thread, threadTable )

	if _threadStack == nil then
		_threadStack = NewStack()
	end
	_threadStack:push( thread )
	_activeThread = thread

	local status, info = coroutine.resume( thread )
	if not status then
		if isDebugging then
			print( info )
		else
			DebugMessage({ Text = info })
		end
		assert( status, info )
	end
	local wait = info and info.wait
	local screenTime = info and info.screenTime
	local event = info and info.event

	_threadStack:pop()
	_activeThread = _threadStack:peek()

	if status and wait ~= nil then
		if wait > 0 then

			local resumeTime
			if screenTime then
				resumeTime = wait + _screenTime
			else
				resumeTime = wait + _worldTime
			end

			table.insert( threadTable, { resumeTime = resumeTime, screenTime = screenTime, thread = thread, tag = info.tag } )
			return "wait"
		elseif wait < 0 then
			assert( event )
			local eventListener = _eventListeners[event]
			if eventListener == nil then
				eventListener = {}
			end
			table.insert( eventListener, thread )
			_eventListeners[event] = eventListener
			return "waitUntil"
		end
	end

	return "done"

end

function hurryUpWaitingThreads( tag )
	--debugprint( "hurryUpWaitingThreads("..tag..")" )
	for k,v in pairs( _threads ) do
		if tag == v.tag then
			if v.screenTime then
				v.resumeTime = _screenTime
			else
				v.resumeTime = _worldTime
			end
		end
	end
end

function dispatch( func )
	local triggerArgs = global_triggerArgs or {}
	local co = coroutine.create( function () func( triggerArgs ) end )

	if verboseLogging then
		debug.sethook( co, newFunctionCall, "cr" )
	end
	local status = resume( co, _threads )

end

-- This will notify an event *and* store the fact that the event has already notified
-- If this event could ever fire more than once, consider Using "notifyExistingWaiters"
function notify( event )
	local eventListeners = _eventListeners[event]
	if eventListeners ~= nil then
		for index, thread in pairs( eventListeners ) do
			resume( thread, _workingThreads )
		end
	else
		-- no one is waiting, store it in case someone asks
		_events[event] = true
	end
	_eventListeners[event] = nil

end

function notifyExistingWaiters( event )
	--debugprint( "notifyExistingWaiters("..event..")" )
	local eventListeners = _eventListeners[event]
	if eventListeners ~= nil then
		_eventListeners[event] = nil
		for index, thread in pairs( eventListeners ) do
			resume( thread, _workingThreads )
		end
	end
end

function thread( func, ... )
	local args = {...};
	local co = coroutine.create( function () func( table.unpack(args) ) end )

	if verboseLogging then
		debug.sethook( co, newFunctionCall, "cr" )
	end
	local status = resume( co, _threads )

end

function update( time, screenTime )

	_worldTime = time
	_screenTime = screenTime

	for k,v in pairs( _threads ) do

		local checkTime = _worldTime
		if v.screenTime then
			checkTime = _screenTime
		end

		if coroutine.status( v.thread ) ~= "dead" and v.resumeTime < checkTime then
			resume( v.thread, _workingThreads )
		else
			table.insert( _workingThreads, v )
		end

	end

	_threads = _workingThreads
	_workingThreads = {}

end

--[[
	LoadMap needs to yield since the actual load doesn't happen until the next frame
	and we need to prevent following statements from executing until the map is loaded.
	C/C# functions can't yield (gives a metamethod/C-call boundary erro).  Instead, LoadMap
	is actually defined in Main.lua which calls LuaMapInternal and then wait(0.01).
]]--
function LoadMap( argTable )
	PickNextMapLoadHint( argTable.Name )
	LoadMapInternal( argTable )
	wait( 0.01 )
end


function eat_true( t, ... )
	assert(t, ...)
	return ...
end

function printData( data, len )
	if len > #data then
		len = #data
	end

	local count = len-100
	for i = 1,count,100 do
		local e = i+100
		local str = "["..i.."-"..e.."]: "
		for j = i,e do
			str = str..data:byte(j,j).." "
		end
		debugprint( str )
	end

	local str = "["..count.."-"..len.."]: "
	for i = count,len do
		str = str..data:byte(i,i).." "
	end
	debugprint( str )
end

SaveIgnores = ToLookup({
	"debug",
	"package",
	"luanet",
	"_G",
	"os",
	"coroutine",
	"table",
	"_currentLine",
	"_currentFileName",
	"_currentStackLevel",
	"lastGoodThreadInfo",
	"string",
	"math",
	"io",
	"_VERSION",
	"Pickle",
	"_threadStack",
	"_threads",
	"_workingThreads",
	"_events",
	"bit32",
	"_eventListeners",
	"error",
	"pcall",
	"rawequal",
	"tostring",
	"getmetatable",
	"setmetatable",
	"rawlen",
	"rawget",
	"loadfile",
	"next",
	"rawset",
	"tonumber",
	"xpcall",
	"print",
	"type",
	"select",
	"dofile",
	"setmetatable",
	"collectgarbage",
	"assert",
	"pairs",
	"ipairs",
	"load",
	"resume",
	"char",
	"newFunctionCall",
	"NewStack",
	"SetCurrentLine",
	"TO_SAVE",
	"luabins",
	"utf8",
	"DebugFunctionIgnores",
	"MainFileFunctions",
	"global_triggerArgs",
	"SaveIgnores",
	"_saveData",

	-- Game Data - Objects which should be considered const by the rest of the scripts
	"Color",
	"NeverFormat",
	"LockerRoomData",
	"NarrationData",
	"DialogueData",
	"SpeechSet",
	"PlayerData",
	"ArchetypeData",
	"PlayerDefaultAttributes",
	"TeamAIs",
	"WorldData",
	"NodeData",
	"PointsOfInterest",
	"ItemData",
	"GuideBookData",
	"InteractiveForageItems",
	"LootDropTables",
	"ConstellationStarData",
	"ArchjusticeAppearanceData",
	"MatchEvents",
	"PlayerTraitsTable",
	"PlayerSkillsTable",
	"PlayerSkillTrees",
	"BendyData",
	"DialoguePositions",
	"speakingCharacterProperties",
	"nonSpeakingCharacterProperties",
	"OptionLetters",
	"ResourceNames",
	"FlyingAtmosphereGroups",
	"FlyingOvermapStates",
	"SkillProcData",
	"LoopFixData",

	"AuraStatTable",
	"SpeedStatTable",
	"RespawnStatTable",

	"ControlComponents",
	"PortraitComponents",
	"StaminaComponents",
	"OvermapComponents",

	"LocationMenuOptionComponents",
	"LocationMenuOtherComponents",
	"LocationMenuComponents",

	"SwathData",
	"SwathElements",
	"InteractPointIds",
	"StormWallIds",
	"SwathCaravanGroups",
	"SwathZeroSpecialSwathLand",
	"SwathZeroSpecialSwathAtmosphere",
	"SkyTable",
	"StarsTable",
	"GlowTable",
	"ScribeStarLinearExceptions",
	"ScribeStarOpenExceptions",
	"OvermapAtmosphereStates",

	"LegalEncounterMaps",
	"Keywords",
	"SeasonSchedule",
	"ImpSwaths",

	"MatchObjectives",
	"XPTable",
	"DifficultyModifiers",

	"Clickables",
	"StarMapObjects",
	"DialoguePositionArchetypeOffsets",
	"IntroMatchDialoguePositionArchetypeOffsets",
	"StarAlignmentLocations",
	"NighttimeOffAssets",
	"TitanStarGroups",

	"MatchTrackList",
	"CharacterMusicTrackList",
	"SongTrackList",

	"lostCoachIds",
	"titleLogoIds",
	"blackBars",

	-- References - Objects regenerated on load so as not to duplicate data -- pointers are broken into serpate objects on save/load
	-- Most of these come from CreateLeagueReferences()
	"PlayerTeam",
	"NextTeam",
	"LastTeam",
	"PrevailedTeam",
	"DefeatedTeam",
	"ForfeitedTeam",
	"RosterScreenTeam",

	"CharBard",
	"CharRivalBard",
	"CharShopKeeper",

	"CharFriend",
	"CharBadass",
	"CharJoker",
	"CharPsycho",
	"CharKnight",
	"CharVamp",
	"CharDoc",
	"CharDiva",
	"CharMascot",

	"CharExplorer",
	"CharForager",
	"CharSeeker",
	"CharDowner",
	"CharProtag",
	"CharLeader",
	"CharTraveler",
	"CharLucky",

	"CharNum1",
	"CharNum2",
	"CharNum3",
	"CharNum4",
	"CharNum5",
	"CharNum6",
	"CharNum7",
	"CharNum8",
	"CharNum9",

	"CharPick1",
	"CharPick2",
	"CharPick3",

	"CharBanished",
	"CharBuddy",

	"CharBuddyPair1",
	"CharBuddyPair2",

	"RivalAccuser",
	"RivalFateDad",
	"RivalFateSon",
	"RivalDissident",

	"RivalWithdrawn",
	"RivalPyreheart",
	"RivalEssence",
	"RivalChastity",
	"RivalTemper",
	"RivalNightwing",
	"RivalBeyonder",

	"RivalTarget",
	"RivalWanderer",
	"RivalPrevailed",
	"RivalDefeated",
	"RivalSeasonFoe",
	"RivalFlying",

	"LowestScorerAllTime",
	"LowestScorerLastMatch",
	"HighestScorerAllTime",
	"HighestScorerLastMatch",
	"OverlookedCharacter",
	"LastAssignedChar",
	"ChallengeCharacter",

	"CharacterTags",
	"currentNodeData",

	-- Game object ids created from SpawnObstacle({ }) that must be destroyed on a load and the object re-created
	"sun",
	"caravanControlsSetup",
	"ObjectiveAnchorIds",

	-- Temp Globals
	"TooltipData",
	"IdsTable",
	"TeamA",
	"TeamB",
	"Char",
	"Chars",
	"Rival",
	"Rivals",
	"RivalTeams",
	"AllCharacters",
	"CharacterPositions",
	"GroupWithSubGroupsTable",
	"IdsByTypeTable",
	"SelectedChapter",
	"SelectedItem",
	"SelectedCharacter",
	"LastAscendedCharacter",
	"LastAscendedRival",
	"LootedItem",
	"TriggerChar",
	"TerminalPressable",
	"Narration",
	"PrevNarration",
	"Standings",
	"ItemStore",
	"CurrentScenePositions",
	"PrevScenePositions",
	"CurrentScenePositionTrackers",

	"MusicId",
	"MusicName",
	"PausedMusicId",
	"PausedMusicName",
	"EndingMusicId",
	"AmbienceId",
	"AmbienceId2",
	"currentAmbience",
	"currentAmbience2",
	"OvermapMusicId",
	"BardMusicId",
	"BeyonderMusicId",

	"LocalizationData",

	-- Optimizations
	"CharacterCache",
	"LocalMP",
})

function Save()

	-- Remove from saving
	if currentDate ~= nil then
		currentDate.String = nil
		currentDate.ShortString = nil
		currentDate.Suffix = nil
	end

	local saveTable = {}
	for key, value in pairs( _G ) do

		if value ~= nil and not SaveIgnores[key] then
			local valueType = type(value)
			if valueType ~= "function" and valueType ~= "userdata" and valueType ~= "thread" then
				if valueType == "table" then
					ValidateTypes( value, key )
					--DebugPrint({ Text = "Saving: "..key.." = "..tostring(value) })
					--local size = CalcTotalNumEntries( value )
					--DebugPrint({ Text = "Saving: "..key.." (NumEntries = "..size..")", LogOnly = true })
				end

				saveTable[key] = value
			end
		end

	end

	_saveData = assert( luabins.save( saveTable ) )

	-- Regenerate after saving
	GenerateDateStrings()

end

function GenerateDateStrings()

	if currentDate ~= nil then
			
		local lang = GetLanguage({ })

		-- date suffix
		if currentDate.Date == 1 or currentDate.Date == 21 or currentDate.Date == 31 then
			currentDate.Suffix = GetDisplayName({ Text = "sufSt" })
		elseif currentDate.Date == 2 or currentDate.Date == 22 then
			currentDate.Suffix = GetDisplayName({ Text = "sufNd" })
		elseif currentDate.Date == 3 or currentDate.Date == 23 then
			currentDate.Suffix = GetDisplayName({ Text = "sufRd" })
		else
			currentDate.Suffix = GetDisplayName({ Text = "sufTh" })
		end

		-- regenerate the date string
		local day = GetDisplayName({ Text = currentDate.Day })
		local month = GetDisplayName({ Text = currentDate.Month })

		-- english first. ShortString/String may need to be changed for certain languages
		-- Leaned on this: https://en.wikipedia.org/wiki/Date_format_by_country  and google translate for German, French & Chinese
		currentDate.String = day..", the "..currentDate.Date..currentDate.Suffix.." of "..month..", "..currentDate.Year.." A.S."
		currentDate.ShortString = day..", "..month.." "..currentDate.Date..GetDisplayName({ Text = currentDate.Suffix })

		if lang == "es" then
			currentDate.String = day..", "..currentDate.Date.." de "..month..", "..currentDate.Year.." A.S."
		elseif lang == "fr" then
			currentDate.String = day..", le "..currentDate.Date.." du "..month..", "..currentDate.Year.." A.S."
		elseif lang == "de" then
			currentDate.String = day..", die "..currentDate.Date.." des "..month..", "..currentDate.Year.." A.S."
		elseif lang == "zh-CN" then
			currentDate.String = currentDate.Year.." "..month.." "..day..", "..currentDate.Date..currentDate.Suffix
		elseif lang == "ru" then
			currentDate.String = day..", "..currentDate.Date..currentDate.Suffix.." "..month..", "..currentDate.Year.." A.S."
		end
		
	end

end

function ValidateTypes( table, tableName )

	-- Strip any non-primitive types from tables before saving

	for tableKey, tableValue in pairs( table ) do
		if type(tableValue) == "table" then
			ValidateTypes( tableValue )
		elseif type(tableValue) == "number" or type(tableValue) == "boolean" or type(tableValue) == "string" then
			-- Fine primitive type
		else
			table[tableKey] = nil
			--[[
			if tableName ~= nil then
				DebugPrint({ Text = "Table: "..tableName, LogOnly = true })
			end
			DebugPrint({ Text = "Stripping non-primitive: "..tableKey.." = "..tostring(tableValue), LogOnly = true })
			--]]
		end
	end

end

function DebugLoad( key, value )
	local t = type(value)
	if t == "boolean" then
		debugprint( "restoring "..key.." = "..(value and "true" or "false") )
	elseif t == "table" then
		debugprint( "restoring "..key.." = "..pickle(value) )
	elseif t == nil then
		debugprint( "restoring "..key.." = nil" )
	else
		debugprint( "restoring "..key.." = "..value )
	end
end

function Load( data, persisted )
	local savedValues = eat_true( luabins.load( data ) )
	for key,value in pairs( savedValues ) do

		--if persisted == nil or persisted[key] then
			--DebugLoad( key, value )
			_G[key] = value
		--end

	end
end


-- http://snippets.luacode.org/snippets/stack_97
-- Stack (for Lua 5.1)
function NewStack( t )

	local Stack =
	{
		push = function( self, ... )
			for _, v in ipairs{...} do
				self[#self+1] = v
			end
		end,

		pop = function( self, num )
			local num = num or 1
			if num > #self then
				error("underflow in NewStack-created stack")
			end
			local ret = {}
			for i = num, 1, -1 do
				ret[#ret+1] = table.remove(self)
			end
			return table.unpack(ret)
		end,

		peek = function( self )
			if #self == 0 then
				return nil
			else
				return self[#self]
			end
		end
	}

	return setmetatable( t or {}, {__index = Stack} )

end

----------------------------------------------
-- Pickle.lua
-- A table serialization utility for lua
-- Steve Dekorte, http://www.dekorte.com, Apr 2000
-- Freeware
----------------------------------------------

function pickle( t )
	return Pickle:clone():pickle_(t)
end

Pickle =
{
	clone = function (t) local nt={}; for i, v in pairs(t) do nt[i]=v end return nt end
	}

	function Pickle:pickle_( root )

		if type(root) ~= "table" then
			error("can only pickle tables, not ".. type(root).."s")
		end

		self._tableToRef = {}
		self._refToTable = {}
		local savecount = 0
		self:ref_(root)
		local s = ""

		for index, value in pairs( root ) do
			if type(value) ~= "function" then
				s = string.format( "%s[%s]=%s,\n", s, self:value_(index), self:value_(value) )
			end
		end

		return string.format("{%s}", s)

	end

	function Pickle:value_( value )

		local vtype = type( value )

		if vtype == "string" then
			return string.format("%q", value)
		elseif vtype == "number" then
			return value
		elseif vtype == "boolean" then
			if value then
				return "true"
			else
				return "false"
			end
		elseif vtype == "table" then
			return self:pickle_(value)
		else
			--error("pickle a "..type(value).." is not supported")
		end

	end

	function Pickle:ref_(t)
		local ref = self._tableToRef[t]
		if not ref then

			if t == self then
				error("can't pickle the pickle class")
			end

			table.insert(self._refToTable, t)
			ref = #self._refToTable
			self._tableToRef[t] = ref

		end
		return ref
	end

----------------------------------------------
-- unpickle
----------------------------------------------

	function unpickle(s)
		if type(s) ~= "string" then
			error("can't unpickle a "..type(s)..", only strings")
		end
		local gentables = load("return "..s)
		local tables = gentables()

		for tnum = 1, #tables do
			local t = tables[tnum]
			local tcopy = {}; for i, v in pairs(t) do tcopy[i] = v end
			for i, v in pairs(tcopy) do
				local ni, nv
				if type(i) == "table" then ni = tables[i[1]] else ni = i end
				if type(v) == "table" then nv = tables[v[1]] else nv = v end
				t[i] = nil
				t[ni] = nv
			end
		end
		return tables[1]
	end

----------------------------------------------
-- PickleTest.lua
-- Testing code for Pickle.lua
-- Steve Dekorte, http://www.dekorte.com, Apr 2000
----------------------------------------------

--dofile("Pickle.lua")

--[[
function test()
  local t = {
	name = "foo",
	ssn=123456789,
	contact = { phone = "555-1\r\n212", email = "foo@foo.com"},
  }
  t.t = { 1 }
  t.contact.loop = t
  t["a b"] = "zzz"
  t[10] = 11
  t[t] = 5
  t[t.t] = 10

  local s = pickle(t)
  print("pickled string:\n\n"..s)

  local ut = unpickle(s)
  print("pickled string:\n\n"..pickle( ut ))
  print("loop test:   "); eq(ut == ut.contact.loop)
  print("subitem test:"); eq(ut.contact.phone == t.contact.phone)
  print("number value:"); eq(ut.ssn == t.ssn)
  print("number index:"); eq(ut[10] == 11)
  print("table index: "); eq(ut[ut] == 5)
end
--]]

	function eq(b)
		if b then print(" succeeded") else print(" failed") end
	end

-- Function Serialization
-- http://lua-users.org/lists/lua-l/2009-11/msg00533.html

	function char( c )
		return ("\\%03d"):format(c:byte())
	end

	function serializeString( s )
		return ('"%s"'):format(s:gsub("[^ !#-~]", char))
	end

-- Split a string into a list of string with sep as seperator
	function string_split(str, sep)
		local sep, fields = sep or ":", {}
		local pattern = string.format("([^%s]+)", sep)
		str:gsub(pattern, function(c) fields[#fields+1] = c end)
		return fields
	end

-- Returns a new table with the same contents as the passed table.
	function shallow_copy(t)
		local t2 = {}
		for k,v in pairs(t) do
			t2[k] = v
		end
		return t2
	end

	function Using( usingName )
	end

	function print_r( t )

		local print_r_cache={}
		local function sub_print_r(t,indent)
			if (print_r_cache[tostring(t)]) then
				print(indent.."*"..tostring(t))
			else
				print_r_cache[tostring(t)]=true
				if (type(t)=="table") then
					for pos,val in pairs(t) do
						if (type(val)=="table") then
							print(indent.."["..pos.."] => "..tostring(t).." {")
							sub_print_r(val,indent..string.rep(" ",string.len(pos)+8))
							print(indent..string.rep(" ",string.len(pos)+6).."}")
						elseif (type(val)=="string") then
							print(indent.."["..pos..'] => "'..val..'"')
						else
							print(indent.."["..pos.."] => "..tostring(val))
						end
					end
				else
					print(indent..tostring(t))
				end
			end
		end
		if (type(t)=="table") then
			print(tostring(t).." {")
			sub_print_r(t,"  ")
			print("}")
		else
			sub_print_r(t,"  ")
		end
		print()
	end